@@ -1,6 +1,6 @@
 ;;; org-mem-updater.el --- Incremental caching -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2025 Free Software Foundation, Inc.
+;; Copyright (C) 2025-2026 Free Software Foundation, Inc.
 
 ;; This program is free software: you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -32,142 +32,84 @@
 (require 'cl-lib)
 (require 'subr-x)
 (require 'llama)
-(require 'el-job-ng)
+(require 'el-job)
 (require 'org-mem)
 (require 'org-mem-parser)
+(define-obsolete-variable-alias 'org-mem-updater--timer 'org-mem-updater--reset-timer "0.28.0 (2026-02-07)")
 
 
 ;;; Targeted-scan
 
-;; FIXME: I think it'd be best not to advise rename-file nor delete-file.
-;; Take e.g. the case where you rename a large number of files in dired...
-;; potentially spawning 1 scan process for each?
-;;
-;; Instead, use `directory-files-and-attributes' like in:
-;; https://github.com/org-roam/org-roam/pull/2558
-;; I.e. re-scan files with changed mtime, remove goners (including renamed)
-;; and scan new files (including renamed), all in one go.
-;;
-;; It'd be triggered after a short timer, or if pseudo-synchrony is important,
-;; triggered after the likes of `dired-do-delete' (or if dired-async used,
-;; `dired-async-message-function'), or an interactive call of `delete-file',
-;; but certainly not on just `delete-file' called from Lisp.
-;; And equivalents for save, write, rename &c.
+(defun org-mem-updater-update (&optional synchronous)
+  "Update cache for each file that has changed, appeared or disappeared.
 
-(defun org-mem-updater--handle-rename (file newname &rest _)
-  "Arrange to scan NEWNAME for entries and links, and forget FILE."
-  (org-mem-updater--handle-delete file)
-  ;; HACK: This function causes `org-mem--rebuild-specially-indexed-tables' to
-  ;; be called twice, making `org-mem--target<>old-links' less useful for
-  ;; downstream use such as `org-node-backlink--maybe-fix-proactively'.
-  ;; Set the table manually to work around that.
-  (let ((tbl (copy-hash-table org-mem--target<>links)))
-    (unless (memq 'move-file-to-trash
-                  (cl-loop for i from 1 to 15 collect (cadr (backtrace-frame i))))
-      (cl-assert newname) ;; b/c below func would accept nil
-      (org-mem-updater--handle-save newname)
-      (setq org-mem--target<>old-links tbl))))
-
-(defun org-mem-updater--handle-delete (file &optional _trash)
-  "Forget entries and links in FILE.
-
-If FILE differs from the name by which the actual file is listed in our
-tables, because a parent directory is a symlink or the abbreviation
-differs, try to discover the known name variant, then forget the data
-from that file.
-
-However, do not do so when FILE itself satisfies `file-symlink-p'.
-In that case, there may be nothing wrong with the known name."
-  (when (and (seq-find (##string-suffix-p % file) org-mem-suffixes)
-             ;; Don't accidentally scrub Tramp paths from org-id-locations
-             ;; just because we chose to never scan them.
-             (not (file-remote-p file)))
-    (let ((bad (list file))
-          (cached-true (gethash file org-mem--wild-filename<>truename)))
-      (when (and cached-true (not (file-symlink-p file)))
-        (push cached-true bad))
-      (org-mem-updater--forget-file-contents bad)
-      (org-mem--invalidate-file-names bad)
-      (org-mem--rebuild-specially-indexed-tables)
-      (mapc #'clrhash (hash-table-values org-mem--key<>subtable)))))
-
-(defun org-mem-updater--handle-save (&optional file)
-  "Arrange to scan entries and links in FILE, or current buffer file."
-  (unless file
-    (if buffer-file-name
-        (setq file buffer-file-name)
-      (error "org-mem-updater--handle-save: Called with nil argument in %S"
-             (current-buffer))))
-  (when (and (seq-find (##string-suffix-p % file) org-mem-suffixes)
-             (not (backup-file-name-p file)))
-    (org-mem-updater--scan-targeted file)))
-
-(defun org-mem-updater--scan-targeted (file)
-  "Arrange to scan FILE or FILEs."
-  (let ((truenames (thread-last
-                     (ensure-list file)
-                     (seq-keep #'org-mem--truename-maybe)
-                     (seq-uniq)
-                     (seq-filter (##cl-loop for xclude in org-mem-exclude
-                                            never (string-search xclude %))))))
-    (when truenames
-      (el-job-ng-run :inject-vars (append (org-mem--mk-work-vars)
-                                          (el-job-ng-vars org-mem-inject-vars))
-                     :require (append '(org-mem-parser) org-mem-load-features)
-                     :inputs truenames
-                     :funcall-per-input #'org-mem-parser--parse-file
-                     :callback #'org-mem-updater--finalize-targeted-scan))))
+If SYNCHRONOUS, block Emacs until done.
+If SYNCHRONOUS and interrupted by a quit, cancel the update."
+  (let* ((db-files (copy-hash-table org-mem--truename<>metadata))
+         (modified-files
+          (cl-loop for truename in (org-mem--list-files-from-fs)
+                   as attr = (file-attributes truename)
+                   as real-mtime = (and attr (file-attribute-modification-time attr))
+                   as db-mtime = (prog1 (org-mem-file-mtime truename)
+                                   (remhash truename db-files))
+                   when (or (not db-mtime)
+                            (not (time-equal-p db-mtime real-mtime)))
+                   collect truename))
+         (removed-files
+          (hash-table-keys db-files))
+         (all (append removed-files modified-files)))
+    ;; If async job already ongoing, this will cancel that and run a new one
+    (el-job-ng-run
+     :id 'org-mem-updater
+     :inject-vars (append (org-mem--mk-work-vars)
+                          (el-job-ng-vars org-mem-inject-vars))
+     :require (cons 'org-mem-parser org-mem-load-features)
+     :inputs all
+     :funcall-per-input #'org-mem-parser--parse-file
+     :callback #'org-mem-updater--finalize-targeted-scan)
+    (when synchronous
+      (el-job-ng-await-or-die
+       'org-mem-updater 3600
+       (format "Running org-mem-updater-update... (files: %S)" all)))))
 
 (defun org-mem-updater--finalize-targeted-scan (parse-results)
-  "Handle PARSE-RESULTS from `org-mem-updater--scan-targeted'."
+  "Handle PARSE-RESULTS from `org-mem-updater-update'."
   (run-hook-with-args 'org-mem-pre-targeted-scan-functions parse-results)
+  (mapc #'clrhash (hash-table-values org-mem--key<>subtable))
   (let (bad-paths problems)
     (with-current-buffer (setq org-mem-scratch (get-buffer-create " *org-mem-scratch*" t))
       (cl-loop for (bad-path problem file-datum entries links) in parse-results do
                (when bad-path (push bad-path bad-paths))
                (when problem (push problem problems))
-               (org-mem-updater--forget-file-contents (car file-datum))
-               (puthash (car file-datum) file-datum org-mem--truename<>metadata)
-               (run-hook-with-args 'org-mem-record-file-functions file-datum)
+               (let ((file (car file-datum)))
+                 (dolist (entry (gethash file org-mem--truename<>entries))
+                   (remhash (org-mem-entry-id entry) org-mem--id<>entry)
+                   (remhash (org-mem-entry-title-maybe entry) org-mem--title<>id)
+                   (remhash (org-mem-entry-pseudo-id entry) org-mem--pseudo-id<>links)
+                   (run-hook-with-args 'org-mem--forget-entry-functions entry))
+                 (remhash file org-mem--truename<>entries)
+                 (remhash file org-mem--truename<>metadata)
+                 (run-hook-with-args 'org-mem--forget-file-functions file) ; Can deprecate
+                 (puthash file file-datum org-mem--truename<>metadata)
+                 (run-hook-with-args 'org-mem--record-file-functions file-datum))
                (dolist (entry entries)
                  (org-mem--record-entry entry)
-                 (run-hook-with-args 'org-mem-record-entry-functions entry))
+                 (run-hook-with-args 'org-mem--record-entry-functions entry))
                (dolist (link links)
-                 (push link (gethash (org-mem-link--internal-entry-id link)
-                                     org-mem--internal-entry-id<>links))
-                 (run-hook-with-args 'org-mem-record-link-functions link))))
-    ;; REVIEW: Assuming it is fine to invalidate now and no need to
-    ;;         move this line to before the above loop
+                 (push link (gethash (org-mem-link-entry-pseudo-id link)
+                                     org-mem--pseudo-id<>links))
+                 (run-hook-with-args 'org-mem--record-link-functions link))))
+    ;; REVIEW: Here we assume it is fine to invalidate now rather than before
+    ;; or during the above loop.  I'm not 100% sure it is.
+    ;; The reason to do it now: it's more efficient when called once on a list.
     (org-mem--invalidate-file-names bad-paths)
     (org-mem--rebuild-specially-indexed-tables)
 
     (run-hook-with-args 'org-mem-post-targeted-scan-functions parse-results)
-    (when bad-paths
-      (let ((good-paths (seq-difference (seq-keep #'org-mem--truename-maybe bad-paths)
-                                        bad-paths)))
-        (when good-paths
-          (org-mem-updater--scan-targeted good-paths))))
     (when problems
       (setq org-mem--problems (append problems org-mem--problems))
       (message "Scan had problems, see M-x org-mem-list-problems"))))
 
-(defun org-mem-updater--forget-file-contents (files)
-  "Delete from tables, most info relating to FILES and their contents.
-You should also run `org-mem--invalidate-file-names'
-and `org-mem--rebuild-specially-indexed-tables'."
-  (setq files (ensure-list files))
-  (when files
-    (with-current-buffer (setq org-mem-scratch (get-buffer-create " *org-mem-scratch*" t))
-      (dolist (file files)
-        (dolist (entry (gethash file org-mem--truename<>entries))
-          (remhash (org-mem-entry-id entry) org-mem--id<>entry)
-          (remhash (org-mem-entry-title-maybe entry) org-mem--title<>id)
-          (remhash (org-mem-entry--internal-id entry) org-mem--internal-entry-id<>links)
-          (run-hook-with-args 'org-mem-forget-entry-functions entry))
-        (remhash file org-mem--truename<>entries)
-        (remhash file org-mem--truename<>metadata)
-        (run-hook-with-args 'org-mem-forget-file-functions file)))))
-
 
 ;;; Instant placeholders
 
@@ -225,9 +167,9 @@ No support for citations."
                             org-mem-exclude))
         (org-mem-updater-ensure-buffer-file-known)
         (let ((link (org-mem-updater-mk-link-atpt)))
-          (push link (gethash (org-mem-link--internal-entry-id link)
-                              org-mem--internal-entry-id<>links))
-          (run-hook-with-args 'org-mem-record-link-functions link))))))
+          (push link (gethash (org-mem-link-entry-pseudo-id link)
+                              org-mem--pseudo-id<>links))
+          (run-hook-with-args 'org-mem--record-link-functions link))))))
 
 (defun org-mem-updater-ensure-id-node-at-point-known ()
   "Record ID-node at point.
@@ -260,7 +202,11 @@ No support for citations."
       (let ((type (org-element-property :type el))
             (desc-beg (org-element-property :contents-begin el))
             (desc-end (org-element-property :contents-end el))
-            (truename (file-truename buffer-file-name)))
+            (truename (file-truename buffer-file-name))
+            (entry-text (buffer-substring (if (org-before-first-heading-p)
+                                              (point-min)
+                                            (org-entry-beginning-position))
+                                          (org-entry-end-position))))
         (record 'org-mem-link
                 truename
                 (point)
@@ -269,11 +215,9 @@ No support for citations."
                 (and desc-beg (buffer-substring-no-properties desc-beg desc-end))
                 nil
                 (org-entry-get-with-inheritance "ID")
-                nil ;; HACK: supplement field is nil
-                (+ (org-mem-parser--hash truename)
-                   (if (org-before-first-heading-p)
-                       0
-                     (org-entry-beginning-position)))))
+                nil ;; HACK: the supplement field is nil
+                (org-mem-parser--mk-id (file-attributes truename)
+                                       entry-text)))
     (error "No link at point %d in %s" (point) (current-buffer))))
 
 (defun org-mem-updater-mk-entry-atpt ()
@@ -291,7 +235,11 @@ Some fields are incomplete or left at nil."
          (scheduled (cdr (assoc "SCHEDULED" properties)))
          (ftitle (org-get-title))
          (title (or heading ftitle))
-         (truename (file-truename buffer-file-name)))
+         (truename (file-truename buffer-file-name))
+         (entry-text (buffer-substring (if (org-before-first-heading-p)
+                                           (point-min)
+                                         (org-entry-beginning-position))
+                                       (org-entry-end-position))))
     (when title
       (setq title (org-link-display-format (substring-no-properties title))))
     (record 'org-mem-entry
@@ -327,7 +275,8 @@ Some fields are incomplete or left at nil."
             (org-mem-updater--tags-at-point-inherited-only)
             (org-get-tags nil t)
             (when heading (org-get-todo-state))
-            (+ (org-mem-parser--hash truename) (if heading pos 0)))))
+            (org-mem-parser--mk-id (file-attributes truename)
+                                   entry-text))))
 
 (defun org-mem-updater--tags-at-point-inherited-only ()
   "Like `org-get-tags', but get only the inherited tags."
@@ -344,22 +293,43 @@ Some fields are incomplete or left at nil."
 
 ;;; Mode
 
-(defvar org-mem-updater--timer (timer-create)
+(defvar org-mem-updater--reset-timer (timer-create)
   "Timer for intermittently running `org-mem--scan-full'.")
 
-(defun org-mem-updater--adjust-timer (&rest _)
-  "Adjust `org-mem-updater--timer' based on duration of last full scan.
+(defun org-mem-updater-adjust-reset-timer (&rest _)
+  "Adjust `org-mem-updater--reset-timer' based on duration of last full scan.
 If timer not running, start it.
 Override this if you prefer different timer delays, or no timer."
   (let ((new-delay (* 10 (1+ org-mem--time-elapsed))))
-    (when (or (not (member org-mem-updater--timer timer-idle-list))
+    (when (or (not (member org-mem-updater--reset-timer timer-idle-list))
               ;; Don't enter an infinite loop -- idle timers can be a footgun.
               (not (> (float-time (or (current-idle-time) 0))
                       new-delay)))
-      (cancel-timer org-mem-updater--timer)
-      (setq org-mem-updater--timer
+      (cancel-timer org-mem-updater--reset-timer)
+      (setq org-mem-updater--reset-timer
             (run-with-idle-timer new-delay t #'org-mem--scan-full)))))
 
+(defun org-mem-updater--update-file-soon (&optional file &rest _)
+  "Schedule to run `org-mem-updater-update' very soon.
+
+Designed for `after-save-hook' and as advice for `delete-file' and
+`rename-file'.  Such functions might be called many times in a loop,
+and this design is meant to avoid invoking `org-mem-updater-update'
+for every FILE, but wait and do a massed invocation afterwards."
+  (setq file (or file buffer-file-name))
+  (when (and file (cl-some (##string-suffix-p % file) org-mem-suffixes))
+    (org-mem-updater--update-soon)))
+
+(defvar org-mem-updater--debounce-timer nil)
+(defun org-mem-updater--update-soon ()
+  "Schedule to run `org-mem-updater-update' very soon.
+If already scheduled, postpone to very soon."
+  (if (memq org-mem-updater--debounce-timer timer-list)
+      (timer-set-time org-mem-updater--debounce-timer
+                      (time-add (current-time) 0.5))
+    (setq org-mem-updater--debounce-timer
+          (run-with-timer 0.5 nil #'org-mem-updater-update))))
+
 ;;;###autoload
 (define-minor-mode org-mem-updater-mode
   "Keep Org-mem cache up to date."
@@ -368,20 +338,21 @@ Override this if you prefer different timer delays, or no timer."
   (require 'org-mem-updater)
   (if org-mem-updater-mode
       (progn
-        (add-hook 'org-mem-post-full-scan-functions #'org-mem-updater--adjust-timer 90)
-        (add-hook 'after-save-hook                  #'org-mem-updater--handle-save)
-        (advice-add 'rename-file :after             #'org-mem-updater--handle-rename)
-        (advice-add 'delete-file :after             #'org-mem-updater--handle-delete)
-        (org-mem-updater--adjust-timer)
+        (add-hook 'org-mem-post-full-scan-functions #'org-mem-updater-adjust-reset-timer 90)
+        (add-hook 'after-save-hook                  #'org-mem-updater--update-file-soon)
+        (advice-add 'rename-file :after             #'org-mem-updater--update-file-soon)
+        (advice-add 'delete-file :after             #'org-mem-updater--update-file-soon)
+        (org-mem-updater-adjust-reset-timer)
         (org-mem--scan-full))
-    (remove-hook 'org-mem-post-full-scan-functions #'org-mem-updater--adjust-timer)
-    (remove-hook 'after-save-hook                  #'org-mem-updater--handle-save)
-    (advice-remove 'rename-file                    #'org-mem-updater--handle-rename)
-    (advice-remove 'delete-file                    #'org-mem-updater--handle-delete)
-    (cancel-timer org-mem-updater--timer)))
+    (remove-hook 'org-mem-post-full-scan-functions #'org-mem-updater-adjust-reset-timer)
+    (remove-hook 'after-save-hook                  #'org-mem-updater--update-file-soon)
+    (advice-remove 'rename-file                    #'org-mem-updater--update-file-soon)
+    (advice-remove 'delete-file                    #'org-mem-updater--update-file-soon)
+    (cancel-timer org-mem-updater--reset-timer)))
 
 (defvar org-mem-updater--id-or-ref-target<>old-links :obsolete)
 (defvar org-mem-updater--new-id-or-ref-targets :obsolete)
+(define-obsolete-function-alias 'org-mem-updater--adjust-timer #'org-mem-updater-adjust-reset-timer "0.28.0 (2026-02-07)")
 
 (provide 'org-mem-updater)
 
