@@ -1,6 +1,6 @@
 ;;; org-mem.el --- Fast info from a large number of Org file contents -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2025 Free Software Foundation, Inc.
+;; Copyright (C) 2025-2026 Free Software Foundation, Inc.
 
 ;; This program is free software: you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -19,8 +19,8 @@
 ;; URL:      https://github.com/meedstrom/org-mem
 ;; Created:  2025-03-15
 ;; Keywords: text
-;; Package-Version: 0.27.2
-;; Package-Revision: 0.27.2-0-g6fa947a5872d
+;; Package-Version: 0.29.1
+;; Package-Revision: 0.29.1-0-g3f9c03039684
 ;; Package-Requires: ((emacs "29.1") (el-job "2.7.1") (llama "0.5.0"))
 
 ;;; Commentary:
@@ -41,10 +41,13 @@
 
 ;;; Code:
 
+(define-obsolete-variable-alias 'org-mem--bump-int 'org-mem-internal-version "2026-01-27 (after 0.26.4)")
+(defconst org-mem-internal-version 30 "Not a version number, but bumped sometimes.")
+
 (require 'cl-lib)
 (require 'subr-x)
 (require 'llama)
-(require 'el-job-ng)
+(require 'el-job)
 (require 'org-mem-parser)
 
 (defvar org-id-locations)
@@ -59,9 +62,8 @@
 (declare-function org-id-locations-save "org-id")
 (declare-function org-id-alist-to-hash "org-id")
 (declare-function org-id-hash-to-alist "org-id")
-(define-obsolete-variable-alias 'org-mem-watch-dirs-exclude 'org-mem-exclude "0.13.0 (May 2025)")
 (unless (fboundp 'el-job-ng-vars)
-  (error "Update to el-job 2.7.1+ to use this version of org-mem"))
+  (display-warning 'org-mem "Update to el-job 2.7.1+ to use this version of org-mem"))
 
 (defgroup org-mem nil "Fast info from a large amount of Org file contents."
   :group 'org)
@@ -74,7 +76,7 @@ This makes the raw text available via accessor `org-mem-entry-text'.
 May slow Emacs init if set to t before enabling `org-mem-updater-mode',
 so you might consider setting it afterwards, instead.
 If so, the text would only be available from the first time that
-`org-mem-updater--timer' fires."
+`org-mem-updater--reset-timer' fires."
   :type 'boolean
   :package-version '(org-mem . "0.9.0"))
 
@@ -181,15 +183,15 @@ These features must be discoverable on `load-path'."
   :package-version '(org-mem . "0.21.0"))
 
 (defcustom org-mem-eval-forms nil
-  "Experimental.
-Quoted forms to eval in child process."
+  "Quoted forms to eval once in child process, before starting parse."
   :type '(repeat sexp))
 
 (defcustom org-mem-ignore-regions-regexps
   '(("^[ \t]*:ROAM_REFS:" . "$")
     ("^[ \t]*:BACKLINKS:" . "$")
-    ("^[ \t]*:BACKLINKS:[ \t]*$" . "^[ \t]*:END:[ \t]*$")
-    ("^[ \t]*#\\+begin_src" . "^[ \t]*#\\+end_src")
+    ("^[ \t]*:BACKLINKS:[ \t]*$" . "^[ \t]*:END:[ \t]*$")   ; Cf. `org-clock-drawer-end-re'
+    ("^[ \t]*\\(?:CLOSED\\|DEADLINE\\|SCHEDULED\\):" . "$") ; Cf. `org-planning-line-re'
+    ("^[ \t]*#\\+begin_src" . "^[ \t]*#\\+end_src")         ; Cf. `org-babel-src-block-regexp'
     ("^[ \t]*#\\+begin_example" . "^[ \t]*#\\+end_example")
     ("^[ \t]*#\\+begin_comment" . "^[ \t]*#\\+end_comment")
     ("^[ \t]*#\\+transclude:" . "$"))
@@ -199,22 +201,7 @@ These regions will not be scanned for links nor active timestamps."
   :package-version '(org-mem . "0.23.0"))
 
 (defvar org-mem-scratch nil
-  "Work buffer held current while executing some hooks.
-These are hooks called many times:
-- `org-mem-record-file-functions'
-- `org-mem-record-entry-functions'
-- `org-mem-record-link-functions'
-- `org-mem-forget-file-functions'
-- `org-mem-forget-entry-functions'
-
-This lets a function on these hooks sidestep the performance overhead of
-`with-temp-buffer' or `with-work-buffer', in favor of using the
-already current buffer.  Just erase it first:
-    \(cl-assert (eq (current-buffer) org-mem-scratch))
-    \(erase-buffer)
-
-Buffer is in `fundamental-mode'.  For an Org buffer see function
-`org-mem-org-mode-scratch'.")
+  "Work buffer held current while executing some hooks.")
 
 
 ;;; Basics
@@ -255,13 +242,13 @@ NOTE! A future version may omit the sigil @ in citekeys.")
   "Previous state of `org-mem--target<>links'.
 For downstream use.")
 
-(defvar org-mem--internal-entry-id<>links (make-hash-table :test 'eq)
-  "1:N table mapping internal entry ID to list of `org-mem-link' records.
+(defvar org-mem--pseudo-id<>links (make-hash-table :test 'equal)
+  "1:N table mapping entry pseudo-ID to list of `org-mem-link' records.
 The list represents all links found in that entry,
-but not in its children.")
+but not in the children of that entry.")
 
 (defvar org-mem--key<>subtable (make-hash-table :test 'eq)
-  "Big bag of memoized values, smelling faintly of cabbage.")
+  "Big bag of memoized values, smelling of freshly cut cabbage.")
 
 (define-inline org-mem--table (key subkey)
   "In a table identified by KEY, access value at SUBKEY.
@@ -301,34 +288,32 @@ Note: All tables cleared often, meant for memoizations."
   (type               nil :read-only t :type string-or-nil)
   (target             ""  :read-only t :type string)
   (description        nil :read-only t :type string-or-nil)
-  (citation-p         nil :read-only t :type boolean)
-  (nearby-id          nil :read-only t :type string-or-nil)
+  (citation-p         nil :read-only t :type boolean       :documentation "Whether target is a citekey e.g. @key2 in [cite:@key1;@key2;@key3].")
+  (nearby-id          nil :read-only t :type string-or-nil :documentation "Inherited ID in entry where link found.")
   (supplement         nil :read-only t :type string-or-nil)
-  (-internal-entry-id -1  :read-only t :type integer))
+  (entry-pseudo-id    -1  :read-only t :type integer))
 
 (cl-defstruct org-mem-entry
-  ;; Ordered for easy visual overview of a printed list of objects.
   (file-truename         ""  :read-only t :type string)
-  (lnum                  -1  :read-only t :type integer)
-  (pos                   -1  :read-only t :type integer)
-  (title-maybe           nil :read-only t :type string-or-nil)
-  (level                 -1  :read-only t :type integer)
-  (id                    nil :read-only t :type string-or-nil)
-  ;; Ordered alphabetically.
-  (active-timestamps-int nil :read-only t :type list)
-  (clocks-int            nil :read-only t :type list)
-  (closed-int            nil :read-only t :type integer-or-nil)
+  (lnum                  -1  :read-only t :type integer        :documentation "Line number at heading.")
+  (pos                   -1  :read-only t :type integer        :documentation "Char position at beginning of heading.")
+  (title-maybe           nil :read-only t :type string-or-nil  :documentation "See `org-mem-entry-title'.")
+  (level                 -1  :read-only t :type integer        :documentation "Number of stars in heading, unaffected by `org-odd-levels-only'.")
+  (id                    nil :read-only t :type string-or-nil  :documentation "Value of ID property.")
+  (active-timestamps-int nil :read-only t :type list           :documentation "See `org-mem-entry-active-timestamps'.")
+  (clocks-int            nil :read-only t :type list           :documentation "See `org-mem-entry-clocks'.")
+  (closed-int            nil :read-only t :type integer-or-nil :documentation "See `org-mem-entry-closed'.")
   (crumbs                nil :read-only t :type list)
-  (deadline-int          nil :read-only t :type integer-or-nil)
+  (deadline-int          nil :read-only t :type integer-or-nil :documentation "See `org-mem-entry-deadline'.")
   (priority              nil :read-only t :type string-or-nil)
   (properties-inherited  nil :read-only t :type list)
   (properties-local      nil :read-only t :type list)
-  (scheduled-int         nil :read-only t :type integer-or-nil)
+  (scheduled-int         nil :read-only t :type integer-or-nil :documentation "See `org-mem-entry-scheduled'.")
   (stats-cookies         nil :read-only t :type list)
   (tags-inherited        nil :read-only t :type list)
   (tags-local            nil :read-only t :type list)
   (todo-state            nil :read-only t :type string-or-nil)
-  (-internal-id          -1  :read-only t :type integer))
+  (pseudo-id             -1  :read-only t :type integer        :documentation "Generated by `org-mem-parser--mk-id'.  Not `eq'-safe, so use `=', `eql' or `equal'."))
 
 
 ;;; To find objects to operate on
@@ -347,7 +332,7 @@ perfectly with `org-id-locations'."
     (mapcar #'directory-abbrev-apply (org-mem-all-file-truenames))))
 
 (defun org-mem-all-files ()
-  "All Org files."
+  "Abbreviated truenames of all Org files."
   (with-memoization (org-mem--table 0 'org-mem-all-files)
     (mapcar #'org-mem--fast-abbrev (org-mem-all-file-truenames))))
 
@@ -358,7 +343,8 @@ When in doubt, you should prefer `org-mem-all-files', because
 
 Org-mem uses truenames internally, but if some public function does
 not work with alternative names, that should be considered a bug."
-  (hash-table-keys org-mem--truename<>metadata))
+  (with-memoization (org-mem--table 0 'org-mem--truename<>metadata)
+    (hash-table-keys org-mem--truename<>metadata)))
 
 (defun org-mem-all-entries ()
   "All entries."
@@ -406,7 +392,16 @@ Citations are `org-mem-link' objects that satisfy
 
 (defun org-mem-entry-by-id (id)
   "The entry with :ID: property equal to \(presumed unique) ID."
-  (and id (gethash id org-mem--id<>entry)))
+  (with-memoization (org-mem--table 31 id)
+    (and id (gethash id org-mem--id<>entry))))
+
+(defun org-mem-entry-by-pseudo-id (pseudo-id)
+  "The entry that has internal PSEUDO-ID.
+See `org-mem-parser--mk-id'."
+  (with-memoization (org-mem--table 54 pseudo-id)
+    (cl-loop for entry in (org-mem-all-entries)
+             when (equal pseudo-id (org-mem-entry-pseudo-id entry))
+             return entry)))
 
 (defun org-mem-entry-at-lnum-in-file (lnum file)
   "The entry that is current at line-number LNUM in FILE."
@@ -465,35 +460,32 @@ Citations are `org-mem-link' objects that satisfy
   "List of entries in same order as they appear in FILE, if FILE known.
 The list always contains at least one entry, which
 represents the content before the first heading."
-  (cl-assert (stringp file))
-  (gethash (org-mem--truename-maybe file) org-mem--truename<>entries))
+  (with-memoization (org-mem--table 32 file)
+    (cl-assert (stringp file))
+    (gethash (org-mem--truename-maybe file) org-mem--truename<>entries)))
 
 (defalias 'org-mem-file-entries #'org-mem-entries-in-file)
 
 (defun org-mem-entries-in-files (files)
   "Combined list of entries from all of FILES."
-  (with-memoization (org-mem--table 13 files)
+  (with-memoization (org-mem--table 33 files)
     (cl-loop for file in (delete-dups (mapcar #'org-mem--truename-maybe files))
              when (stringp file)
              append (gethash file org-mem--truename<>entries))))
 
 (defun org-mem-file-by-id (id)
   "The file that contains an :ID: property matching ID."
-  (let ((entry (and id (gethash id org-mem--id<>entry))))
-    (and entry (org-mem-entry-file entry))))
-
-(defun org-mem-entry-that-contains-link (link)
-  "The entry where LINK was found."
-  (org-mem-entry-at-pos-in-file (org-mem-link-file link)
-                                (org-mem-link-pos link)))
+  (with-memoization (org-mem--table 35 file)
+    (let ((entry (and id (gethash id org-mem--id<>entry))))
+      (and entry (org-mem-entry-file entry)))))
 
 (defun org-mem-id-nodes-in-files (files)
   "All ID-nodes in FILES."
   (with-memoization (org-mem--table 15 files)
-    (setq files (delete-dups
-                 (seq-keep #'org-mem--truename-maybe (ensure-list files))))
-    (seq-filter (##member (org-mem-entry-file-truename %) files)
-                (org-mem-all-id-nodes))))
+    (let ((files (delete-dups
+                  (seq-keep #'org-mem--truename-maybe (ensure-list files)))))
+      (seq-filter (##member (org-mem-entry-file-truename %) files)
+                  (org-mem-all-id-nodes)))))
 
 (defun org-mem-links-with-type-and-path (type path)
   "Links with components TYPE and PATH, see `org-link-plain-re'."
@@ -504,13 +496,15 @@ represents the content before the first heading."
 
 (defun org-mem-id-links-to-entry (entry)
   "All ID-links that point to ENTRY."
-  (let ((id (org-mem-entry-id entry)))
-    (and id (org-mem-id-links-to-id id))))
+  (with-memoization (org-mem--table 34 entry)
+    (let ((id (org-mem-entry-id entry)))
+      (and id (org-mem-id-links-to-id id)))))
 
 (defun org-mem-links-to-target (target)
   "All link objects with link target equal to TARGET."
-  (cl-assert (stringp target))
-  (gethash target org-mem--target<>links))
+  (with-memoization (org-mem--table 36 target)
+    (cl-assert (stringp target))
+    (gethash target org-mem--target<>links)))
 
 (defun org-mem-id-links-to-id (id)
   "All ID-links targeting ID."
@@ -523,7 +517,8 @@ represents the content before the first heading."
 
 (defun org-mem-id-node-by-title (title)
   "The ID-node titled TITLE."
-  (and title (gethash (org-mem-id-by-title title) org-mem--id<>entry)))
+  (with-memoization (org-mem--table 37 title)
+    (and title (gethash (org-mem-id-by-title title) org-mem--id<>entry))))
 
 (defun org-mem-id-by-title (title)
   "The ID that currently corresponds to TITLE.
@@ -532,7 +527,8 @@ TITLE is either a heading, a file title, or an alias.
 Assumes unique titles.  If two IDs exist with same title, it is
 undefined which ID is returned.  User can prevent this from becoming a
 problem with the help of option `org-mem-do-warn-title-collisions'."
-  (and title (gethash title org-mem--title<>id)))
+  (with-memoization (org-mem--table 38 title)
+    (and title (gethash title org-mem--title<>id))))
 
 (defun org-mem-links-from-id (id)
   "Links from context where local or inherited ID property is ID."
@@ -568,10 +564,11 @@ problem with the help of option `org-mem-do-warn-title-collisions'."
 
 (defun org-mem-links-in-entry (entry)
   "All links found inside ENTRY, ignoring descendant entries.
-Do not trust the result if used during `org-mem-forget-entry-functions'
+Do not trust the result if used during `org-mem--forget-entry-functions'
 or similar hook.  Trustworthy on `org-mem-post-full-scan-functions'."
-  (and entry (gethash (org-mem-entry--internal-id entry)
-                      org-mem--internal-entry-id<>links)))
+  (with-memoization (org-mem--table 39 entry)
+    (and entry (gethash (org-mem-entry-pseudo-id entry)
+                        org-mem--pseudo-id<>links))))
 
 
 ;;; Entry info
@@ -581,7 +578,8 @@ or similar hook.  Trustworthy on `org-mem-post-full-scan-functions'."
 Better than `org-mem-entry-file-truename' when users may see the name.
 When in doubt, prefer this, but it should not matter what form of file
 name you input to the org-mem API."
-  (org-mem--fast-abbrev (org-mem-entry-file-truename entry)))
+  (with-memoization (org-mem--table 40 entry)
+    (org-mem--fast-abbrev (org-mem-entry-file-truename entry))))
 
 (defun org-mem-entry-subtree-p (entry)
   "Non-nil if ENTRY is a subtree, nil if a \"file-level node\"."
@@ -613,25 +611,25 @@ Excludes text of child entries."
              (mapcar #'cl-fourth (cdr (reverse (org-mem-entry-crumbs entry))))
            (mapcar #'cl-fourth (reverse (org-mem-entry-crumbs entry)))))))
 
-;; TODO: Deprecate the optional arg
 (defun org-mem-entry-olpath-with-file-title (entry &optional filename-fallback)
   "Outline path to ENTRY, including file #+title.
-Argument FILENAME-FALLBACK is deprecated,
+Optional argument FILENAME-FALLBACK is deprecated,
 use `org-mem-entry-olpath-with-file-title-or-basename' instead."
-  (with-memoization (org-mem--table 27 (list entry filename-fallback))
-    (if filename-fallback
-        (org-mem-entry-olpath-with-file-title-or-basename entry)
-      (let ((olp (mapcar #'cl-fourth (reverse (cdr (org-mem-entry-crumbs entry)))))
-            file-name-handler-alist)
-        ;; The car of `olp' is the potentially nil file title.
-        (when (null (car olp))
-          (pop olp))
-        olp))))
+  (when filename-fallback
+    (error "Argument FILENAME-FALLBACK is deprecated"))
+  (with-memoization (org-mem--table 27 entry)
+    (let ((olp (mapcar #'cl-fourth (reverse (cdr (org-mem-entry-crumbs entry)))))
+          file-name-handler-alist)
+      ;; The car of `olp' is the potentially nil file title.
+      (when (null (car olp))
+        (pop olp))
+      olp)))
 
-;; Better than above; unary function.
 (defun org-mem-entry-olpath-with-file-title-or-basename (entry)
   "Outline path to ENTRY, including file #+title.
-Use file basename if there is no #+title."
+Use file basename if there is no #+title.
+
+Basename means `file-name-nondirectory', not `file-name-base'."
   (with-memoization (org-mem--table 30 entry)
     (let ((olp (mapcar #'cl-fourth (reverse (cdr (org-mem-entry-crumbs entry)))))
           file-name-handler-alist)
@@ -642,27 +640,31 @@ Use file basename if there is no #+title."
               olp))
       olp)))
 
-;; TODO: Deprecate the optional arg
 (defun org-mem-entry-olpath-with-self-with-file-title (entry &optional filename-fallback)
   "Outline path, including file #+title, and ENTRY\\='s own heading.
-With FILENAME-FALLBACK, use file basename if there is no #+title.
+Optional argument FILENAME-FALLBACK is deprecated,
+use `org-mem-entry-olpath-with-self-with-file-title-or-basename' instead.
 
-If ENTRY is itself a file-level entry, this still results in a list of
-zero or one strings, not two."
-  (with-memoization (org-mem--table 24 (list entry filename-fallback))
+If ENTRY is itself a file-level entry, the return value is still a list
+of zero or one strings, not two."
+  (when filename-fallback
+    (error "Argument FILENAME-FALLBACK is deprecated"))
+  (with-memoization (org-mem--table 24 entry)
     (let ((olp (mapcar #'cl-fourth (reverse (org-mem-entry-crumbs entry))))
           file-name-handler-alist)
       ;; The car of `olp' is the potentially nil file title.
       (when (null (car olp))
-        (pop olp)
-        (when filename-fallback
-          (push (file-name-nondirectory (org-mem-entry-file-truename entry))
-                olp)))
+        (pop olp))
       olp)))
 
 (defun org-mem-entry-olpath-with-self-with-file-title-or-basename (entry)
   "Outline path, including file #+title, and ENTRY\\='s own heading.
-Use file basename if there is no #+title."
+Use file basename if there is no #+title.
+
+Basename means `file-name-nondirectory', not `file-name-base'.
+
+If ENTRY is itself a file-level entry, the return value is still a list
+of one string, not two."
   (with-memoization (org-mem--table 29 entry)
     (let ((olp (mapcar #'cl-fourth (reverse (org-mem-entry-crumbs entry))))
           file-name-handler-alist)
@@ -676,11 +678,17 @@ Use file basename if there is no #+title."
 (defun org-mem-entry-title (entry)
   "Like `org-mem-entry-title-maybe' but always return a string.
 In the case that ENTRY is a file-level entry with no title, return the
-file basename \(file name sans directory component\)."
-  (or (org-mem-entry-title-maybe entry)
-      (progn (cl-assert (not (org-mem-entry-subtree-p entry)))
-             (let (file-name-handler-alist)
-               (file-name-nondirectory (org-mem-entry-file-truename entry))))))
+file basename \(file name sans directory component\).
+
+I.e. `file-name-nondirectory', not `file-name-base'.  In org-mem jargon,
+\"basename\" refers to what you get from the POSIX command \"basename\"
+with one argument, which should be familiar if you are not too mired in
+Emacs Lisp."
+  (with-memoization (org-mem--table 41 entry)
+    (or (org-mem-entry-title-maybe entry)
+        (progn (cl-assert (not (org-mem-entry-subtree-p entry)))
+               (let (file-name-handler-alist)
+                 (file-name-nondirectory (org-mem-entry-file-truename entry)))))))
 
 (defalias 'org-mem-entry-properties #'org-mem-entry-properties-local
   "Alist of ENTRY properties, no inheritance.
@@ -697,21 +705,24 @@ only the properties explicitly written in the file.")
 
 (defun org-mem-entry-property (prop entry)
   "Value of property PROP in ENTRY."
-  (cdr (assoc (upcase prop) (org-mem-entry-properties entry))))
+  (with-memoization (org-mem--table 42 (list prop entry))
+    (cdr (assoc (upcase prop) (org-mem-entry-properties entry)))))
 
 (defun org-mem-entry-property-with-inheritance (prop entry)
   "Value of property PROP in ENTRY."
   ;; NOTE: A value can be nil (i.e. not string "nil" but symbol nil), but
   ;;       should still override any inherited value, so remember to use
   ;;       `assoc' correctly with that in mind.
-  (cdr (or (assoc (upcase prop) (org-mem-entry-properties entry))
-           (assoc (upcase prop) (org-mem-entry-properties-inherited entry)))))
+  (with-memoization (org-mem--table 43 (list prop entry))
+    (cdr (or (assoc (upcase prop) (org-mem-entry-properties entry))
+             (assoc (upcase prop) (org-mem-entry-properties-inherited entry))))))
 
 (defun org-mem-entry-tags (entry)
   "ENTRY tags, with inheritance if allowed at ENTRY.
 Combines `org-mem-entry-tags-local' and `org-mem-entry-tags-inherited'."
-  (delete-dups (append (org-mem-entry-tags-inherited entry)
-                       (org-mem-entry-tags-local entry))))
+  (with-memoization (org-mem--table 44 entry)
+    (delete-dups (append (org-mem-entry-tags-inherited entry)
+                         (org-mem-entry-tags-local entry)))))
 
 (define-inline org-mem--iso8601 (int-time)
   "Translate INT-TIME into a string \"yyyy-mm-ddThh:mm\"."
@@ -719,28 +730,34 @@ Combines `org-mem-entry-tags-local' and `org-mem-entry-tags-inherited'."
 
 (defun org-mem-entry-closed (entry)
   "CLOSED-timestamp of ENTRY, suitable for `iso8601-parse'."
-  (let ((ts (org-mem-entry-closed-int entry)))
-    (and ts (org-mem--iso8601 ts))))
+  (with-memoization (org-mem--table 45 entry)
+    (let ((ts (org-mem-entry-closed-int entry)))
+      (and ts (org-mem--iso8601 ts)))))
 
 (defun org-mem-entry-deadline (entry)
   "DEADLINE-timestamp of ENTRY, suitable for `iso8601-parse'.
 WARNING: If the timestamp is expressed as a diary-sexp,
 such as <%%\(memq (calendar-day-of-week date) \\='(1 2 3 4 5)))>,
 this returns nil!"
-  (let ((ts (org-mem-entry-deadline-int entry)))
-    (and ts (org-mem--iso8601 ts))))
+  (with-memoization (org-mem--table 46 entry)
+    (let ((ts (org-mem-entry-deadline-int entry)))
+      (and ts (org-mem--iso8601 ts)))))
 
 (defun org-mem-entry-scheduled (entry)
   "SCHEDULED-timestamp of ENTRY, suitable for `iso8601-parse'.
 WARNING: If the timestamp is expressed as a diary-sexp,
 such as <%%\(memq (calendar-day-of-week date) \\='(1 2 3 4 5)))>,
 this returns nil!"
-  (let ((ts (org-mem-entry-scheduled-int entry)))
-    (and ts (org-mem--iso8601 ts))))
+  (with-memoization (org-mem--table 47 entry)
+    (let ((ts (org-mem-entry-scheduled-int entry)))
+      (and ts (org-mem--iso8601 ts)))))
 
 (defun org-mem-entry-active-timestamps (entry)
-  "Active timestamps in ENTRY, suitable for `iso8601-parse'."
-  (mapcar #'org-mem--iso8601 (org-mem-entry-active-timestamps-int entry)))
+  "Active timestamps in ENTRY, suitable for `iso8601-parse'.
+Excludes such timestamps in DEADLINE or SCHEDULED, since there is
+`org-mem-entry-deadline' &c for that."
+  (with-memoization (org-mem--table 48 entry)
+    (mapcar #'org-mem--iso8601 (org-mem-entry-active-timestamps-int entry))))
 
 (defun org-mem-entry-clocks (entry)
   "Alist \((START END MINUTES) ...) representing clock lines in ENTRY.
@@ -749,18 +766,20 @@ MINUTES is an integer amount of minutes.
 To get all three as Lisp time values, use `org-mem-entry-clocks-int'.
 
 Any dangling clock line is represented as just \(START)."
-  (cl-loop for (start end secs) in (org-mem-entry-clocks-int entry)
-           if end collect (list (org-mem--iso8601 start)
-                                (org-mem--iso8601 end)
-                                (/ secs 60))
-           else collect (list (org-mem--iso8601 start))))
+  (with-memoization (org-mem--table 49 entry)
+    (cl-loop for (start end secs) in (org-mem-entry-clocks-int entry)
+             if end collect (list (org-mem--iso8601 start)
+                                  (org-mem--iso8601 end)
+                                  (/ secs 60))
+             else collect (list (org-mem--iso8601 start)))))
 
 (defun org-mem-entry-dangling-clocks (entry)
   "List \(START1 START2 ...) representing clocks in ENTRY with no end.
 See also `org-mem-all-entries-with-dangling-clock'."
-  (cl-loop for clock in (org-mem-entry-clocks-int entry)
-           when (length= clock 1)
-           collect (org-mem--iso8601 (car clock))))
+  (with-memoization (org-mem--table 50 entry)
+    (cl-loop for clock in (org-mem-entry-clocks-int entry)
+             when (length= clock 1)
+             collect (org-mem--iso8601 (car clock)))))
 
 
 ;;; Link info
@@ -768,12 +787,20 @@ See also `org-mem-all-entries-with-dangling-clock'."
 (defun org-mem-link-file (link)
   "Abbreviated truename of file where LINK is.
 See more info at `org-mem-entry-file'."
-  (org-mem--fast-abbrev (org-mem-link-file-truename link)))
+  (with-memoization (org-mem--table 51 link)
+    (org-mem--fast-abbrev (org-mem-link-file-truename link))))
+
+(defun org-mem-link-entry (link)
+  "The entry that contains LINK."
+  (with-memoization (org-mem--table 52 link)
+    (org-mem-entry-at-pos-in-file (org-mem-link-pos link)
+                                  (org-mem-link-file link))))
 
 (defun org-mem-id-link-p (link)
   "Return t if LINK is an `org-mem-link' object of type \"id\"."
-  (and (org-mem-link-p link)
-       (equal (org-mem-link-type link) "id")))
+  (with-memoization (org-mem--table 53 link)
+    (and (org-mem-link-p link)
+         (equal (org-mem-link-type link) "id"))))
 
 
 ;;; File info
@@ -951,7 +978,7 @@ file has been scanned, but in practice they may often be identical."
 (defun org-mem-roam-reflink-p (link)
   "Non-nil if target of LINK is known to exist in some ROAM_REFS.
 Relies on up-to-date table `org-mem--roam-ref<>id', so do not trust it
-during the hook `org-mem-forget-entry-functions' or similar."
+during the hook `org-mem--forget-entry-functions' or similar."
   (and (org-mem-link-p link)
        (gethash (org-mem-link-target link) org-mem--roam-ref<>id)))
 
@@ -995,8 +1022,8 @@ These link-targets are determined by `org-mem--split-roam-refs-field'."
 
 (defun org-mem-roam-reflinks-into-file (file)
   "Reflinks from anywhere, leading into somewhere in FILE."
-  (mapcan #'org-mem-roam-reflinks-to-entry
-          (org-mem-entries-in-file file)))
+  (seq-mapcat #'org-mem-roam-reflinks-to-entry
+              (org-mem-entries-in-file file)))
 
 (defun org-mem-roam-reflinks-into-files (files)
   "Reflinks from anywhere, leading into somewhere in FILES."
@@ -1073,10 +1100,10 @@ What is valid?  See \"org-mem-test.el\"."
                    ;; .. but the actual ref is just the //path
                    path))))))
 
-(add-hook 'org-mem-record-entry-functions
+(add-hook 'org-mem--record-entry-functions
           #'org-mem--record-roam-aliases-and-refs -10)
 
-(add-hook 'org-mem-forget-entry-functions
+(add-hook 'org-mem--forget-entry-functions
           #'org-mem--forget-roam-aliases-and-refs -10)
 
 
@@ -1141,13 +1168,19 @@ What is valid?  See \"org-mem-test.el\"."
   (:method ((xx org-mem-entry)) (org-mem-entry-file-truename xx))
   (:method ((xx org-mem-link)) (org-mem-link-file-truename xx)))
 
+(cl-defgeneric org-mem-pseudo-id (entry/link)
+  "Pseudo-ID of ENTRY/LINK.
+See `org-mem-parser--mk-id'."
+  (:method ((xx org-mem-entry)) (org-mem-entry-pseudo-id xx))
+  (:method ((xx org-mem-link)) (org-mem-link-entry-pseudo-id xx)))
+
 (cl-defgeneric org-mem-id (entry/file)
   "ID property of ENTRY/FILE - if file name, the file-level ID."
   (:method ((xx org-mem-entry)) (org-mem-entry-id xx))
   (:method ((xx string)) (org-mem-file-id-strict xx)))
 
-(cl-defgeneric org-mem-title (entry-or-file)
-  "Heading title, or file title or basename of ENTRY-OR-FILE.
+(cl-defgeneric org-mem-title (file/entry)
+  "Heading title, or file title or basename of ENTRY/FILE.
 Like `org-mem-entry-title', this always returns a string."
   (:method ((xx org-mem-entry)) (org-mem-entry-title xx))
   (:method ((xx string)) (org-mem-file-title-or-basename xx)))
@@ -1176,13 +1209,13 @@ Like `org-mem-entry-title', this always returns a string."
 (defvar org-mem-post-full-scan-functions nil
   "Hook passed the list of parse-results, after a full reset.")
 
-(defvar org-mem-record-file-functions nil
+(defvar org-mem--record-file-functions nil
   "Hook passed (FILE ATTRS LINES PTMAX) after adding that info to tables.")
 
-(defvar org-mem-record-entry-functions nil
+(defvar org-mem--record-entry-functions nil
   "Hook passed one `org-mem-entry' object after adding it to tables.")
 
-(defvar org-mem-record-link-functions nil
+(defvar org-mem--record-link-functions nil
   "Hook passed one `org-mem-link' object after adding it to tables.")
 
 (defvar org-mem-pre-targeted-scan-functions nil
@@ -1195,10 +1228,10 @@ hence the name.  Contrast `org-mem-pre-full-scan-functions'.")
 This occurs after scanning a targeted single file or set of files,
 hence the name.  Contrast `org-mem-post-full-scan-functions'.")
 
-(defvar org-mem-forget-file-functions nil
+(defvar org-mem--forget-file-functions nil
   "Hook passed a file name after removing its info from tables.")
 
-(defvar org-mem-forget-entry-functions nil
+(defvar org-mem--forget-entry-functions nil
   "Hook passed one forgotten `org-mem-entry' object.")
 
 (defvar org-mem--problems nil)
@@ -1216,6 +1249,10 @@ For arguments TAKEOVER and MSG, see `org-mem--scan-full'."
   (interactive "i\ni\np")
   (push (list (format-time-string "%FT%T.%3N") takeover msg called-interactively)
         org-mem--resets-debug-log)
+  ;; Temporary because we break downstream in 123d77a
+  (when (and (featurep 'org-node)
+             (not (boundp 'org-node-internal-version)))
+    (display-warning 'org-mem "You probably want to update org-node, not just org-mem"))
   (when called-interactively
     (setq org-mem--next-message t)
     (setq takeover t))
@@ -1281,7 +1318,7 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
   (clrhash org-mem--id<>entry)
   (clrhash org-mem--truename<>metadata)
   (clrhash org-mem--truename<>entries)
-  (clrhash org-mem--internal-entry-id<>links)
+  (clrhash org-mem--pseudo-id<>links)
   (setq org-mem--title-collisions nil)
   (let (bad-paths problems)
 
@@ -1291,14 +1328,14 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
                (when bad-path (push bad-path bad-paths))
                (when problem (push problem problems))
                (puthash (car file-data) file-data org-mem--truename<>metadata)
-               (run-hook-with-args 'org-mem-record-file-functions file-data)
+               (run-hook-with-args 'org-mem--record-file-functions file-data)
                (dolist (entry entries)
                  (org-mem--record-entry entry)
-                 (run-hook-with-args 'org-mem-record-entry-functions entry))
+                 (run-hook-with-args 'org-mem--record-entry-functions entry))
                (dolist (link links)
-                 (push link (gethash (org-mem-link--internal-entry-id link)
-                                     org-mem--internal-entry-id<>links))
-                 (run-hook-with-args 'org-mem-record-link-functions link))))
+                 (push link (gethash (org-mem-link-entry-pseudo-id link)
+                                     org-mem--pseudo-id<>links))
+                 (run-hook-with-args 'org-mem--record-link-functions link))))
     (org-mem--invalidate-file-names bad-paths)
     (org-mem--rebuild-specially-indexed-tables)
 
@@ -1375,7 +1412,7 @@ remove, it's easiest to wipe and re-build."
   (setq org-mem--target<>old-links (copy-hash-table org-mem--target<>links))
   (clrhash org-mem--target<>links)
   (cl-loop
-   for links being each hash-value of org-mem--internal-entry-id<>links
+   for links being each hash-value of org-mem--pseudo-id<>links
    do (dolist (link links)
         (push link (gethash (org-mem-link-target link) org-mem--target<>links)))))
 
@@ -1687,7 +1724,7 @@ As an obsolete calling convention, MESSAGE can also be a symbol
 corresponding to your package name."
   (when (symbolp message)
     (setq message (format "%S waiting for org-mem..." message)))
-  (el-job-ng-await 'org-mem-targeted n-secs message)
+  (el-job-ng-await 'org-mem-updater n-secs message)
   (el-job-ng-await 'org-mem n-secs message))
 
 ;; REVIEW: Mixed feelings about including this tool, but it's the obvious tool
@@ -1745,11 +1782,11 @@ This is a transition helper.  Org-mem 0.27.0 changed the style of
 PARSE-RESULTS passed to `org-mem-pre-full-scan-functions' & co.
 
 For reference, the new style is a list as long as the number of files,
-each element being a list:
+each element being a list of 5 lists:
 
-   \(BAD-PATH PROBLEM FILE-DATUM ENTRIES LINKS)
+   \((BAD-PATH PROBLEM FILE-DATUM ENTRIES LINKS) ...)
 
-Old style was a list of 5 lists:
+Old style was single list of 5 lists:
 
    \(BAD-PATHS FILE-DATA ENTRIES LINKS PROBLEMS)"
   (cl-loop
@@ -1799,8 +1836,8 @@ org-id-locations:
       ;; Bonus, probably unnecessary
       (dolist (file files)
         (dolist (entry (gethash file org-mem--truename<>entries))
-          (remhash (org-mem-entry--internal-id entry)
-                   org-mem--internal-entry-id<>links))
+          (remhash (org-mem-entry-pseudo-id entry)
+                   org-mem--pseudo-id<>links))
         (remhash file org-mem--truename<>entries)
         (remhash file org-mem--truename<>metadata))
       (setq org-id-files (cl-nset-difference org-id-files files :test #'equal))
@@ -1809,9 +1846,6 @@ org-id-locations:
       (org-mem--scan-full))))
 
 
-(defvaralias 'org-mem-internal-version 'org-mem--bump-int) ; Alias added 2026-01-27 (after 0.26.3)
-(defconst org-mem--bump-int 24 "Not a version number, but bumped sometimes.")
-
 (defmacro org-mem--def-whiny-alias (old new when removed-by)
   "Define function OLD as effectively an alias for NEW.
 Also, calling OLD will emit a deprecation warning the first time.
@@ -1826,13 +1860,37 @@ may be removed from the package."
                 ,when ,removed-by ,new))
        (apply ,new args))))
 
-(defvar org-mem--file<>metadata :renamed)
-(defvar org-mem--file<>entries  :renamed)
-
+(defconst org-mem-watch-dirs-exclude :renamed-0.13.0)
 (org-mem--def-whiny-alias 'org-mem-entry-olpath-with-file-title-with-self #'org-mem-entry-olpath-with-self-with-file-title  "0.22.0 (2025-10-01)" "2026-04-30")
 (org-mem--def-whiny-alias 'org-mem-olpath-with-file-title-with-self       #'org-mem-olpath-with-self-with-file-title        "0.22.0 (2025-10-01)" "2026-04-30")
+(defconst org-mem-forget-link-functions  :obsolete-0.24.0)
 
-(defvar org-mem-forget-link-functions :obsolete) ;; 2025-10-15 but unused for long already
+(defconst org-mem-forget-entry-functions :obsolete-0.29.0
+  "Deprecated because dangerous.
+Still exists under name `org-mem--forget-entry-functions',
+but please use `org-mem-post-targeted-scan-functions'.")
+
+(defconst org-mem-forget-file-functions  :obsolete-0.29.0
+  "Deprecated because dangerous.
+Still exists under name `org-mem--forget-file-functions',
+but please use `org-mem-post-targeted-scan-functions'.")
+
+(defconst org-mem-record-link-functions  :obsolete-0.29.0
+  "Deprecated because dangerous.
+Still exists under name `org-mem--record-link-functions',
+but please use `org-mem-post-full-scan-functions'.")
+
+(defconst org-mem-record-entry-functions :obsolete-0.29.0
+  "Deprecated because dangerous.
+Still exists under name `org-mem--record-entry-functions',
+but please use `org-mem-post-full-scan-functions'.")
+
+(defconst org-mem-record-file-functions  :obsolete-0.29.0
+  "Deprecated because dangerous.
+Still exists under name `org-mem--record-file-functions',
+but please use `org-mem-post-full-scan-functions'.")
+
+(define-obsolete-function-alias 'org-mem-entry-that-contains-link #'org-mem-link-entry "0.29.0 (2026-02-11)")
 
 (provide 'org-mem)
 
